import termios
import fcntl
import struct
import os
import sys

colors = dict(red=31, green=32, yellow=33, blue=34, cyan=36, white=37)


def get_terminal_width():
    try:
        call = fcntl.ioctl(1, termios.TIOCGWINSZ, "\000" * 8)
        width = struct.unpack("hhhh", call)[1]
    except:
        width = int(os.environ.get('COLUMNS', 80))
    return width if width >= 40 else 80


def colored(text, color='white'):
    text = '\x1b[{0}m{1}\x1b[0m'.format(colors.get(color, 37), text)
    sys.stdout.write(text)


def separated(text, color='red', sepchar='='):
    fullwidth = get_terminal_width()
    if sys.platform == "win32":
        fullwidth -= 1
    n = (fullwidth - len(text) - 2) // (2 * len(sepchar))
    fill = sepchar * n
    line = "{} {} {}".format(fill, text, fill)
    if len(line) + len(sepchar.rstrip()) <= fullwidth:
        line += sepchar.rstrip()
    colored(line, color)



from .search import search
from .parsers import py_parser
from .config import conf
from .utils import report


def check():
    files = search(conf.EXTENSIONS, conf.EXCLUDE, conf.BASE_DIR)
    parsed = parse_files(files['.py'], py_parser)
    unused, maybe_unused = find_unused(
        parsed['defined_objects'],
        parsed['used_objects']
    )
    report(unused, maybe_unused)


def parse_files(paths, parser):
    result = {}
    for path in paths:
        parsed = parser.parse(path)
        for key, value in parsed.items():
            result.setdefault(key, set()).update(value)
    return result


def find_unused(defined, used):
    maybe_unused = find_duplicate_endings(defined, used)
    unused = [item for item in defined if not item.endswith(tuple(used))]
    return unused, maybe_unused


def find_duplicate_endings(defined, used):
    overall = {}
    maybe_unused = []
    for used_item in used:
        for defined_item in defined:
            if defined_item.endswith(used_item):
                overall.setdefault(used_item, set()).add(defined_item)

    for _, value in overall.items():
        if len(value) > 1:
            maybe_unused.append(value)
    return maybe_unused



import sys
import os

from .config import conf
from .writer import colored, separated

ERRORS = {
    1: "Syntax error in file {0}: {1}\n",
    2: "Unable to detect unused names, 'from {0} import *' used in file {1}.\n"
}


def error(code, str_args):
    assert isinstance(str_args, list)
    colored(ERRORS[code].format(*str_args), 'red')
    sys.exit()


def get_dot_relpath(path):
    cutted_path, _ = os.path.splitext(path)
    relpath = os.path.relpath(cutted_path, conf.BASE_DIR)
    return relpath.replace('/', '.')


def report(unused, maybe_unused):
    if unused:
        separated('UNUSED PYTHON CODE')
        for item in sorted(unused, key=lambda x: (x.lower(), x.line)):
            colored('- {}:{}: '.format(item.filepath, item.line), 'cyan')
            colored('Unused {} "{}"\n'.format(item.node, item.name), 'yellow')
    else:
        separated('NO UNUSED PYTHON CODE', 'green')

    if maybe_unused:
        colored('\n\nIt is recommended to check the next groups of items, they may be unused:\n',
                'white')
        for group in maybe_unused:
            separated('-', 'white', '-')
            for item in group:
                colored('- {}:{}: '.format(item.filepath, item.line), 'white')
                colored('{} "{}"\n'.format(item.node, item.name), 'yellow')
        separated('-', 'white', '-')


class ParsedItem(str):

    def __new__(cls, path, name, filepath, node, line, module=None):
        item = str.__new__(cls, path)
        item.name = name
        item.filepath = filepath
        item.node = node
        item.line = line
        item.module = module
        return item



import ast

from .utils import ParsedItem, error, get_dot_relpath


class PyParser(ast.NodeVisitor):

    def reset(self, path):
        self.imported = {}
        self.defined = set()
        self.used = set()
        self.class_child_names = set()
        self.relpath = get_dot_relpath(path)

    def parse(self, path):
        self.reset(path)
        ast_node = self.get_ast_node(path)
        self.visit(ast_node)
        return self.prepare_results()

    @staticmethod
    def get_ast_node(path):
        module_string = open(path).read()
        try:
            node = ast.parse(module_string, filename=path)
        except SyntaxError as e:
            error(1, [path, e])
        return node

    def prepare_results(self):
        imported_and_used, defined_and_used = self.group_used_items()
        defined_and_used = self.remove_class_child_names(defined_and_used)
        return {
            'defined_objects': self.defined - defined_and_used,
            'used_objects': imported_and_used
        }

    def group_used_items(self):
        imported, defined = set(), list()
        for item in self.used:
            if item.node == 'attribute':
                if item.module in self.imported:
                    imported.add('{0}.{1}'.format(self.imported[item.module], item.name))
                    continue
            if item.name in self.imported:
                imported.add(self.imported[item.name])
            else:
                defined.append(item)
        return imported, defined

    def remove_class_child_names(self, names):
        used = set()
        for name in names:
            if '{}_{}'.format(name.name, name.line) not in self.class_child_names:
                used.add(name)
        return used

    def get_relpath_from_import(self, node):
        if node.level != 0:
            pathlist = self.relpath.split('.')[:-node.level]
            if node.module:
                pathlist.append(node.module)
            return '.'.join(pathlist)
        return node.module

    def get_item(self, node, node_type, module=None):
        name = self.get_name_value(node)
        node_path = self.get_node_path(name)
        return ParsedItem(node_path, name, self.relpath, node_type, node.lineno, module)

    def get_node_path(self, name):
        return '{0}.{1}'.format(self.relpath, name)

    @staticmethod
    def get_name_value(node):
        name = getattr(node, 'name', None)
        node_id = getattr(node, 'id', None)
        attr = getattr(node, 'attr', None)
        return name or node_id or attr

    def visit(self, node):
        method = 'visit_' + node.__class__.__name__.lower()
        visitor = getattr(self, method, None)
        if visitor is not None:
            visitor(node)
        return self.generic_visit(node)

    def visit_classdef(self, node):
        if node.col_offset == 0:
            self.defined.add(self.get_item(node, 'class'))
            for child in ast.walk(node):
                if isinstance(child, ast.Name) and child.id == node.name:
                    self.class_child_names.add('{}_{}'.format(child.id, child.lineno))

    def visit_attribute(self, node):
        if isinstance(node.ctx, ast.Load) and isinstance(node.value, ast.Name):
            if node.value.id != 'self':
                self.used.add(self.get_item(node, 'attribute', node.value.id))

    def visit_functiondef(self, node):
        if node.col_offset == 0:
            self.defined.add(self.get_item(node, 'function'))

    def visit_name(self, node):
        self.used.add(self.get_item(node, 'name'))

    def visit_import(self, node):
        for item in node.names:
            name = item.asname or item.name
            self.imported[name] = name

    def visit_importfrom(self, node):
        module_path = self.get_relpath_from_import(node)
        for item in node.names:
            name = item.asname or item.name
            if name == '*':
                error(2, [node.module, self.relpath])
            self.imported[name] = '{0}.{1}'.format(module_path, item.name)

py_parser = PyParser()



from setuptools import setup

setup(
    name='pydead',
    version='1.1.1',
    author='Sergey Petrov',
    author_email='srgypetrov@ya.ru',
    url='https://github.com/SrgyPetrov/pydead',
    description='Utility for searching of unused code in python projects',
    packages=['src'],
    scripts=['pydead'],
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Topic :: Utilities',
    ]
)



import os

from fnmatch import fnmatchcase


def search(extensions, exclude, base_path):

    fix_extensions(extensions)
    files = {extension: list() for extension in extensions}

    def scan_directory(path):
        for item in os.listdir(path):
            itempath = os.path.join(path, item)

            if os.path.isfile(itempath):
                ext = get_ext_or_none(itempath)
                if ext:
                    files[ext].append(itempath)
            else:
                scan_directory(itempath)

    def get_ext_or_none(filepath):
        _, extension = os.path.splitext(filepath)
        exclude_match = [fnmatchcase(filepath, pattern) for pattern in exclude]
        if extension not in extensions or any(exclude_match):
            return None
        return extension

    scan_directory(base_path)
    return files


def fix_extensions(extensions):
    for i, ext in enumerate(extensions):
        if not ext.startswith('.'):
            extensions[i] = '.' + ext



import os
import configparser


base_dir = os.getcwd()
config_file = os.path.join(base_dir, '.pydead')

parser = configparser.ConfigParser()
parser.read(config_file)


class AttributeDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


conf = AttributeDict(
    EXCLUDE=['.*'] + parser.get('paths', 'exclude', fallback='').split('\n'),
    EXTENSIONS=['.py'],
    BASE_DIR=base_dir
)
